// NDS GML Thingy
// This is a little application that lets you scribble
// on the DS and save the result as a GML file

// Copyright 2010 Aaron Gray <aaronwgray@gmail.com>
// Distributed under GNU GPL (v2 and v3)

/*
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <nds.h>
#include <stdio.h>
#include <vector>
#include <fat.h>
#include <time.h>

using namespace std;

// color defines, only b+w are currently used
// 16 bit color, 1 alpha + 3x5 color
// binary format is abbbbbgggggrrrrr
#define COLOR_BLACK 0x8000
#define COLOR_WHITE 0xFFFF
#define COLOR_GREY 0xC210
#define COLOR_RED 0x801F
#define COLOR_GREEN 0x83E0
#define COLOR_BLUE 0xFC00
#define COLOR_TEAL 0xFFE0
#define COLOR_PINK 0xFC1F
#define COLOR_YELLOW 0x83FF

// type definitions -- a graffiti contains strokes which contains points
struct point {
	int x;
	int y;
	int t;
};
typedef vector<point> stroke;
typedef vector<stroke> graffiti;

// global variables
stroke G_stroke;
graffiti G_graffiti;
volatile int G_ticks;
touchPosition G_touch;
volatile int G_touch_x;
volatile int G_touch_y;
bool G_file_lock;
uint16 G_color_fg;
uint16 G_color_bg;
uint16* G_screen_buffer;

// redraws the screen buffer based on G_graffiti
void redraw_buffer() {
    dmaFillHalfWords( G_color_bg, G_screen_buffer, 256*192*2);
	for( graffiti::const_iterator s = G_graffiti.begin(); s != G_graffiti.end(); s++ ) {
		for( stroke::const_iterator p = s->begin(); p != s->end(); p++ ) {
			G_screen_buffer[(p->x-1) + 256 * (p->y-1)] = G_color_fg;
		}
	}

}

// saves a gml file based on G_graffiti
void save_gml() {
	G_file_lock = true;

	time_t utime = time(NULL);
	struct tm* ltime = localtime((const time_t *)&utime);

	// open a new file, increment file name counter
	char filepath[32];
	int first_tick = G_graffiti.begin()->begin()->t;
	sprintf( filepath, "fat:/dsgml/%04d%02d%02d%02d%02d%02d.gml",
		ltime->tm_year+1900, ltime->tm_mon, ltime->tm_mday,
		ltime->tm_hour, ltime->tm_min, ltime->tm_sec);
	FILE* gmlfile = fopen (filepath, "w");
	if( gmlfile == NULL ) {
		return;
	}

	// boilerplate gml header info
	fprintf(gmlfile, "<gml spec='0.1c'>\n");
	fprintf(gmlfile, "<tag>\n");
	fprintf(gmlfile, "<header>\n");
	fprintf(gmlfile, "<environment>\n");
	fprintf(gmlfile, "<up>\n");
	fprintf(gmlfile, "<x>0.0</x>\n");
	fprintf(gmlfile, "<y>-1.0</y>\n");
	fprintf(gmlfile, "<z>0.0</z>\n");
	fprintf(gmlfile, "</up>\n");
	fprintf(gmlfile, "<screenBounds>\n");
	fprintf(gmlfile, "<x>256</x>\n");
	fprintf(gmlfile, "<y>192</y>\n");
	fprintf(gmlfile, "<z>0</z>\n");
	fprintf(gmlfile, "</screenBounds>\n");
	fprintf(gmlfile, "</environment>\n");
	fprintf(gmlfile, "</header>\n");
	fprintf(gmlfile, "<drawing>\n");

	// this is where the magic happens
	for( graffiti::const_iterator s = G_graffiti.begin(); s != G_graffiti.end(); s++ ) {
		fprintf(gmlfile, "<stroke>\n");
		for( stroke::const_iterator p = s->begin(); p != s->end(); p++ ) {
			fprintf(gmlfile, "<pt>\n");
			fprintf(gmlfile, "<x>%f</x>\n", float(p->x) / 256);
			fprintf(gmlfile, "<y>%f</y>\n", float(p->y) / 192);
			fprintf(gmlfile, "<t>%f</t>\n", float(p->t - first_tick) / 1000 );
			fprintf(gmlfile, "</pt>\n");
		}
		fprintf(gmlfile, "</stroke>\n");
	}

	// boilerplate gml end tags
	fprintf(gmlfile, "</drawing>\n");
	fprintf(gmlfile, "</tag>\n");
	fprintf(gmlfile, "</gml>\n");

	fclose(gmlfile);

	G_file_lock = false;
}

// reads and responds to button and stylus input
void handle_input() {
	scanKeys();
	touchRead(&G_touch);

	// moving the stylus updates the global stroke
	if( keysHeld() & KEY_TOUCH ) {
		if(G_touch_x != G_touch.px || G_touch_y != G_touch.py) {
			G_touch_x = G_touch.px;
			G_touch_y = G_touch.py;

			point p;
			p.x = G_touch_x;
			p.y = G_touch_y;
			p.t = G_ticks;

			G_stroke.push_back(p);

			G_screen_buffer[(p.x-1) + 256 * (p.y-1)] = G_color_fg;
		}
	}
	else {
		G_touch_x = 0;
		G_touch_y = 0;
	}

	// lifting the stylus adds the global stroke to the graffiti
	if( keysUp() & KEY_TOUCH ) {
		G_graffiti.push_back(G_stroke);
		G_stroke.clear();
	}

	// START key saves GML file
	if( keysDown() & KEY_START ) {
		save_gml();
	}

	// SELECT key swaps bg/fg colors
	if( keysDown() & KEY_SELECT ) {
		uint16 temp = G_color_bg;
		G_color_bg = G_color_fg;
		G_color_fg = temp;
		redraw_buffer();
	}

	// R key erases previous stroke
	if( keysDown() & KEY_R ) {
		if( G_graffiti.size() ) {
			G_graffiti.pop_back();
			redraw_buffer();
		}
	}

	// L key erases entire graffiti
	if( keysDown() & KEY_L ) {
		G_graffiti.clear();
		redraw_buffer();
	}
}

// vblank interrupt, updates text and copies over screen buffer
void int_vblank() {
	if( G_file_lock ) {
		return;
	}

	handle_input();

	iprintf("\x1b[10;1HTouch X Pos: %03d", G_touch_x);
	iprintf("\x1b[11;1HTouch Y Pos: %03d", G_touch_y);
	iprintf("\x1b[13;1HFrame Count: %08d", G_ticks);
	iprintf("\x1b[15;1HPoint Count: %04d", G_stroke.size());
	iprintf("\x1b[16;1HStroke Count: %04d", G_graffiti.size());

	dmaCopyHalfWords(2, G_screen_buffer, (uint16 *)BG_BMP_RAM_SUB(0), 256*192*2);
}

// timer interrupt, this is where all of the control logic lives
void int_timer() {
	if( G_file_lock ) {
		return;
	}

	G_ticks++;
}

// main function includes setup followed by infinite loop
int main() {

	// initialize global variables to sane values
	G_ticks = 0;
	G_touch.px = 0;
	G_touch.py = 0;

	G_color_fg = COLOR_BLACK;
	G_color_bg = COLOR_WHITE;

	G_file_lock = false;

	G_stroke.clear();
	G_graffiti.clear();

	G_screen_buffer = new uint16[256*192];
    dmaFillHalfWords( G_color_bg, G_screen_buffer, 256*192*2);

	// system & library init
    powerOn(POWER_ALL_2D);
    lcdMainOnTop();
	fatInitDefault();

	// vram setup
	vramSetBankA(VRAM_A_MAIN_BG);
	vramSetBankB(VRAM_B_LCD);
	vramSetBankC(VRAM_C_SUB_BG);
	vramSetBankD(VRAM_D_LCD);

	// main BG is tile mode for text display
    videoSetMode( MODE_5_2D |
                  DISPLAY_BG0_ACTIVE
                );

	// sub BG is bmp mode for graffiti display
    videoSetModeSub( MODE_5_2D |
                     DISPLAY_BG3_ACTIVE
                   );

    REG_BG3CNT_SUB = BG_BMP16_256x256 |
                     BG_BMP_BASE(0) |
                     BG_PRIORITY(3);

    REG_BG3PA_SUB = 1 << 8;
    REG_BG3PB_SUB = 0;
    REG_BG3PC_SUB = 0;
    REG_BG3PD_SUB = 1 << 8;

    REG_BG3X_SUB = 0;
    REG_BG3Y_SUB = 0;

	// console init and base text
	consoleInit(0, 0,BgType_Text4bpp, BgSize_T_256x256, 31,0, true, true);

	iprintf("\x1b[1;1HNDS GML Thingy");
	iprintf("\x1b[3;1HR Clears Last Stroke");
	iprintf("\x1b[4;1HL Clears All Strokes");
	iprintf("\x1b[5;1HStart Saves GML File");
	iprintf("\x1b[6;1HSelect Swaps Color Set");
	iprintf("\x1b[21;1HC 2010 Aaron Gray");
	iprintf("\x1b[22;1HDistributed Under GNU GPLv2/3");

	// register the vblank interrupt
	irqSet(IRQ_VBLANK, int_vblank);

	// initialize and register timer interrupt (1 milisecond resolution)
	TIMER_CR(0) = 0;
	TIMER_DATA(0) = TIMER_FREQ_1024(1000);
	TIMER_CR(0) = TIMER_ENABLE | TIMER_DIV_1024 | TIMER_IRQ_REQ;
	irqSet(IRQ_TIMER0, int_timer);
	irqEnable(IRQ_TIMER0);

	// infinite loop
	while(1) {
		swiWaitForVBlank();
	}

	// just for the hell of it
	return 0;
}